﻿// Copyright (c) 2012 Morten Bakkedal
// This code is published under the MIT License.

using System;
using System.Collections.Generic;

namespace FuncLib.Collections
{
	public class DependencyList<T> where T : class
	{
		private LinkedList<T> items;
		private Dictionary<T, LinkedListNode<T>> nodes;
		private Dictionary<LinkedListNode<T>, ulong> positions;

		/// <summary>
		/// Creates a new empty dependency list.
		/// </summary>
		public DependencyList()
		{
			items = new LinkedList<T>();
			nodes = new Dictionary<T, LinkedListNode<T>>();
			positions = new Dictionary<LinkedListNode<T>, ulong>();
		}

		/// <summary>
		/// Creates a new dependency list with independent items. Dependency relations may be imposed
		/// later by calling <see cref="AddDependency" />.
		/// </summary>
		public DependencyList(IEnumerable<T> items)
			: this()
		{
			if (items == null)
			{
				throw new ArgumentNullException();
			}

			foreach (T item in items)
			{
				Add(item);
			}
		}

		/// <summary>
		/// Adds an independent item to the list. Dependency relations may be imposed by calling <see cref="AddDependency" /> afterwards.
		/// Returns true if dependency item wasn't in the list already (i.e. if the list was enlarged); false otherwise.
		/// </summary>
		public bool Add(T item)
		{
			if (item == null)
			{
				throw new ArgumentNullException();
			}

			if (nodes.ContainsKey(item))
			{
				// Item already in list.
				return false;
			}

			LinkedListNode<T> node = items.AddLast(item);
			nodes.Add(item, node);

			// Don't care much about the position here. If needed later, positions will be recalculated automatically.
			positions.Add(node, ulong.MaxValue);

			// New item was added.
			return true;
		}

		/// <summary>
		/// Adds a new dependency to the list if not already added, and make sure that the dependency item is before
		/// the current item. Returns true if dependency item wasn't in the list already (i.e. if the list was enlarged);
		/// false otherwise.
		/// </summary>
		public bool AddDependency(T item, T dependencyItem)
		{
			if (item == null || dependencyItem == null)
			{
				throw new ArgumentNullException();
			}

			LinkedListNode<T> node;
			if (!nodes.TryGetValue(item, out node))
			{
				throw new ArgumentException("Can't add relative to an item not in the list.", "item");
			}

			LinkedListNode<T> dependencyNode;
			if (nodes.TryGetValue(dependencyItem, out dependencyNode))
			{
				if (ComparePositions(node, dependencyNode) < 0)
				{
					// Moving is needed. Remove from the list. 
					items.Remove(dependencyNode);

					// Add again at the right position.
					AddDependency(node, dependencyItem);
				}

				// Item was already in the list.
				return false;
			}
			else
			{
				// Not in list. Add just before current item. (Or in top of the list?)
				AddDependency(node, dependencyItem);

				// New item was added.
				return true;
			}
		}

		private void AddDependency(LinkedListNode<T> node, T dependencyItem)
		{
			// Add to the linked list.
			LinkedListNode<T> dependencyNode = items.AddBefore(node, dependencyItem);

			// Compute "pseudo" position of the new dependency.
			ulong p1;
			if (dependencyNode.Previous != null)
			{
				p1 = positions[dependencyNode.Previous];
			}
			else
			{
				p1 = 0;
			}
			ulong p2 = positions[node];

			// Generate position between the two positions above: p1 <= p <= p2 is always satisfied.
			ulong p = p1 / 2 + p2 / 2;
			positions[dependencyNode] = p;

			// Add for fast lookup in the linked list.
			nodes[dependencyItem] = dependencyNode;
		}

		private int ComparePositions(LinkedListNode<T> node1, LinkedListNode<T> node2)
		{
			if (node1.Value == node2.Value)
			{
				// Skip test if items are equal by reference.
				return 0;
			}

			int compare = positions[node1].CompareTo(positions[node2]);

			if (compare == 0)
			{
				// Identical "pseudo" positions. Rescan and equidistantial
				UpdatePositions();

				// Compare again.
				compare = positions[node1].CompareTo(positions[node2]);
			}

			return compare;
		}

		private void UpdatePositions()
		{
			ulong p = ulong.MaxValue;
			ulong delta = ulong.MaxValue / (ulong)items.Count;

			for (LinkedListNode<T> node = items.Last; node != null; node = node.Previous)
			{
				positions[node] = p;
				p -= delta;
			}
		}

		public T[] ToArray()
		{
			// Copy the sorted structure to a more compact representation in an array.
			T[] array = new T[items.Count];
			items.CopyTo(array, 0);

			return array;
		}

		public int Count
		{
			get
			{
				return items.Count;
			}
		}

		/*public class Node
		{
			public T Item
			{
				get;
				set;
			}

			public ulong Position
			{
				get;
				set;
			}
		}
		
		private class HalfLinkedList
		{
			public void AddBefore(Node node, T value)
			{
				Node node2 = new Node();
				node2.Value = value;
				node2.

			}
		}

		private class Node
		{
			public T Value
			{
				get;
				set;
			}

			public Node Previous
			{
				get;
				set;
			}

			public ulong Position
			{
				get;
				set;
			}
		}*/
	}
}
