﻿// Copyright (c) 2012 Morten Bakkedal
// This code is published under the MIT License.

using System;
using System.Collections.Generic;

namespace FuncLib.Functions
{
	public static class Evaluator
	{
		public static IEvaluator Create(IPoint point)
		{
			return new InnerEvaluator(point);
		}

		public static IEvaluator CreateCompact(IPoint point)
		{
			return new InnerCompactEvaluator(point);
		}

		public static IPartialEvaluator CreatePartial(IPoint point)
		{
			// It's important for efficiency that the same objects are reused whenever possible. Thus no compact memory
			// implementation of IPartialEvaluator (like InnerCompactEvaluator; but in that case only primitive types
			// are returned, i.e. doubles).
			return new InnerPartialEvaluator(point);
		}

		public static double Evaluate(IPoint point, Function function)
		{
			return Evaluate(point, new Function[] { function })[0];
		}

		public static double[] Evaluate(IPoint point, Function[] functions)
		{
			return Evaluate(Create(point), functions);
		}

		public static double EvaluateCompact(IPoint point, Function function)
		{
			return EvaluateCompact(point, new Function[] { function })[0];
		}

		public static double[] EvaluateCompact(IPoint point, Function[] functions)
		{
			return Evaluate(CreateCompact(point), functions);
		}

		public static Function PartialEvaluate(IPoint point, Function function)
		{
			return CreatePartial(point).PartialValue(function);
		}

		private static double[] Evaluate(IEvaluator evaluator, Function[] functions)
		{
			List<double> values = new List<double>();
			foreach (Function function in functions)
			{
				values.Add(evaluator.Value(function));
			}

			return values.ToArray();
		}

		private class InnerEvaluator : IEvaluator
		{
			private Dictionary<Function, double> values0;
			private Dictionary<object, object> values;

			public InnerEvaluator(IPoint point)
			{
				Point = point;

				values0 = new Dictionary<Function, double>();
				values = new Dictionary<object, object>();
			}

			public double Value(Function function)
			{
				double value;
				if (!values0.TryGetValue(function, out value))
				{
					value = Function.ComputeValue(function, this);
					values0.Add(function, value);
				}

				return value;
			}

			public void Add<TKey, TValue>(TKey key, TValue value)
			{
				values[key] = value;
			}

			public bool TryGetValue<TKey, TValue>(TKey key, out TValue value)
			{
				object value2;
				if (values.TryGetValue(key, out value2) && value2 is TValue)
				{
					value = (TValue)value2;
					return true;
				}

				value = default(TValue);
				return false;
			}

			public IPoint Point
			{
				get;
				private set;
			}
		}

		private class InnerCompactEvaluator : IEvaluator
		{
			public InnerCompactEvaluator(IPoint point)
			{
				Point = point;
			}

			public double Value(Function function)
			{
				// Don't store function values.
				return Function.ComputeValue(function, this);
			}

			public void Add<TKey, TValue>(TKey key, TValue value)
			{
				// Don't store values.
			}

			public bool TryGetValue<TKey, TValue>(TKey key, out TValue value)
			{
				value = default(TValue);
				return false;
			}

			public IPoint Point
			{
				get;
				private set;
			}
		}

		private class InnerPartialEvaluator : IPartialEvaluator
		{
			private Dictionary<Function, Function> partialValues;

			public InnerPartialEvaluator(IPoint point)
			{
				Point = point;

				partialValues = new Dictionary<Function, Function>();
			}

			public Function PartialValue(Function function)
			{
				Function partialValue;
				if (!partialValues.TryGetValue(function, out partialValue))
				{
					partialValue = Function.ComputePartialValue(function, this);
					partialValues.Add(function, partialValue);
				}

				return partialValue;
			}

			public IPoint Point
			{
				get;
				private set;
			}
		}
	}
}
